<?php
/*
  Part of the KD2 framework collection of tools: http://dev.kd2.org/

  Copyright (c) 2001-2016 BohwaZ <http://bohwaz.net/>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
  THE POSSIBILITY OF SUCH DAMAGE.
*/

namespace KD2;

/**
 * Simple error and exception handler
 *
 * When enabled (with ErrorManager::enable(ErrorManager::DEVELOPMENT)) it will
 * catch any error, warning or exception and display it along with useful debug
 * information. If enabled it will also log the errors to a file and/or send
 * every error by email.
 *
 * In production mode no details are given, but a unique reference to the log
 * or email is displayed.
 *
 * This is similar in a way to http://tracy.nette.org/
 *
 * @author  bohwaz <http://bohwaz.net/>
 * @package KD2fw
 * @license BSD
 */
class ErrorManager
{
	/**
	 * Prod/dev modes
	 */
	const PRODUCTION = 1;
	const DEVELOPMENT = 2;

	/**
	 * Term colors
	 */
	const RED = '[1;41m';
	const RED_FAINT = '[1m';
	const YELLOW = '[33m';

	/**
	 * true = catch exceptions, false = do nothing
	 * @var null
	 */
	static protected $enabled = null;

	/**
	 * HTML template used for displaying production errors
	 * @var string
	 */
	static protected $production_error_template = '<!DOCTYPE html><html><head><title>Internal server error</title>
		<style type="text/css">
		body {font-family: sans-serif; }
		code, p, h1 { max-width: 400px; margin: 1em auto; display: block; }
		code { text-align: right; color: #666; }
		a { color: blue; }
		form { text-align: center; }
		input { padding: .3em; }
		</style></head><body><h1>Server error</h1><p>Sorry but the server encountered an internal error and was unable
		to complete your request. Please try again later.</p>
		<if(sent)><p>The webmaster has been noticed and this will be fixed ASAP.</p></if>
		<if(logged)><code>Error reference: <b>{$ref}</b></code></if>
		<if(report)><form method="post" action="{$report_url}"><input type="hidden" value="{$report_json}" /><input type="submit" value="Report this error" /></report_form></if>
		<p><a href="/">&larr; Go back to the homepage</a></p>
		</body></html>';

	/**
	 * E-Mail address where to send errors
	 * @var boolean
	 */
	static protected $email_errors = false;

	/**
	 * Reporting URL
	 */
	static protected $report_url = null;

	/**
	 * Reporting automatically?
	 */
	static protected $report_auto = true;

	/**
	 * Custom context
	 */
	static protected $context = [];

	/**
	 * Custom exception handlers
	 * @var array
	 */
	static protected $custom_handlers = [];

	/**
	 * Does the terminal support ANSI colors
	 * @var boolean
	 */
	static protected $term_color = false;

	/**
	 * Will be set to true when catching an exception to avoid double catching
	 * with the shutdown function
	 * @var boolean
	 */
	static protected $catching = false;

	/**
	 * Used to store timers and memory consumption
	 * @var array
	 */
	static protected $run_trace = [];

	/**
	 * Handles PHP shutdown on fatal error to be able to catch the error
	 * @return void
	 */
	static public function shutdownHandler()
	{
		// Stop here if disabled or if the script ended with an exception
		if (!self::$enabled || self::$catching)
			return false;

		$error = error_get_last();

		if (in_array($error['type'], [E_ERROR, E_CORE_ERROR, E_COMPILE_ERROR, E_PARSE, E_RECOVERABLE_ERROR, E_USER_ERROR], TRUE))
		{
			// Don't exit at the end, as there might be other shutdown handlers
			// after this one
			self::exceptionHandler(new \ErrorException($error['message'], 0, $error['type'], $error['file'], $error['line']), false);
		}
	}

	/**
	 * Internal error handler to throw them as exceptions
	 * (private use)
	 */
	static public function errorHandler($severity, $message, $file, $line)
	{
		if (!(error_reporting() & $severity)) {
			// Don't report this error (for example @unlink)
			return;
		}

		$types = [
			E_ERROR             => 'Fatal error',
			E_USER_ERROR        => 'User error',
			E_RECOVERABLE_ERROR => 'Recoverable error',
			E_CORE_ERROR        => 'Core error',
			E_COMPILE_ERROR     => 'Compile error',
			E_PARSE             => 'Parse error',
			E_WARNING           => 'Warning',
			E_CORE_WARNING      => 'Core warning',
			E_COMPILE_WARNING   => 'Compile warning',
			E_USER_WARNING      => 'User warning',
			E_NOTICE            => 'Notice',
			E_USER_NOTICE       => 'User notice',
			E_STRICT            => 'Strict standards',
			E_DEPRECATED        => 'Deprecated',
			E_USER_DEPRECATED   => 'User deprecated',
		];

		$type = array_key_exists($severity, $types) ? $types[$severity] : 'Unknown error';
		$message = $type . ': ' . $message;

		// Catch ASSERT_BAIL errors differently because throwing an exception
		// in this case results in an execution shutdown, and shutdown handler
		// isn't even called. See https://bugs.php.net/bug.php?id=53619
		if (assert_options(ASSERT_ACTIVE) && assert_options(ASSERT_BAIL) && substr($message, 0, 18) == 'Warning: assert():')
		{
			$message .= ' (ASSERT_BAIL detected)';
			return self::exceptionHandler(new \ErrorException($message, 0, $severity, $file, $line));
		}

		throw new \ErrorException($message, 0, $severity, $file, $line);
		return true;
	}

	/**
	 * Main exception handler
	 * @param  object  $e    Exception or Error (PHP 7) object
	 * @param  boolean $exit Exit the script at the end
	 * @return void
	 */
	static public function exceptionHandler($e, $exit = true)
	{
		self::$catching = true;

		try {
			self::reportException($e, $exit);
			return true;
		}
		catch (\Throwable $e) {
			echo $e;
			exit(1);
		}
		catch (\Exception $e) {
			echo $e;
			exit(1);
		}
	}

	/**
	 * Main exception handler
	 * @param  object  $e    Exception or Error (PHP 7) object
	 * @param  boolean $exit Exit the script at the end
	 * @return void
	 */
	static public function reportException($e, $exit)
	{
		foreach (self::$custom_handlers as $class=>$callback)
		{
			if ($e instanceOf $class)
			{
				call_user_func($callback, $e);
				$e = false;
				break;
			}
		}

		if ($e === false)
		{
			if ($exit)
			{
				exit(1);
			}
			else
			{
				return;
			}
		}

		$report = self::makeReport($e);
		$log = sprintf('=========== Error ref. %s ===========', $report->context->id)
			. PHP_EOL . PHP_EOL . (string) $e . PHP_EOL . PHP_EOL
			. '<errorReport>' . PHP_EOL . json_encode($report, \JSON_PRETTY_PRINT)
			. PHP_EOL . '</errorReport>' . PHP_EOL;

		// Log exception to file
		if (ini_get('error_log'))
		{
			error_log($log);
		}

		// Log exception to email
		if (self::$email_errors)
		{
			// From: sender
			$from = !empty($_SERVER['SERVER_NAME']) ? $_SERVER['SERVER_NAME'] : basename($_SERVER['DOCUMENT_ROOT']);
			mail(self::$email_errors, sprintf('Error #%s: %s', $report->context->id, $e->getMessage()), $log, 'From: ' . sprintf('"%s" <%s>', $from, self::$email_errors));
		}

		unset($e);

		// Disable any output if it was buffering
		if (ob_get_level())
		{
			ob_end_clean();
		}

		if (PHP_SAPI == 'cli')
		{
			foreach ($report->errors as $e)
			{
				self::termPrint(sprintf(' /!\\ %s ', $e->type), self::RED);
				self::termPrint($e->message, self::RED_FAINT);

				if (isset($e->line))
				{
					self::termPrint(sprintf('Line %d in %s', $e->line, $e->file), self::
						YELLOW);
				}

				// Ignore the error stack belonging to ErrorManager
				foreach ($e->backtrace as $i=>$t)
				{
					$file = $t->file ?: '[internal function]';
					$line = $t->line ? '(' . $t->line . ')' : '';

					if (isset($t->args))
					{
						$args = $t->args;

						foreach ($args as &$arg)
						{
							if (strlen($arg) > 20)
							{
								$arg = substr($arg, 0, 19) . '…';
							}
						}

						unset($arg);

						self::termPrint(sprintf('#%d %s%s: %s(%s)', $i, $file, $line, $t->function, implode(', ', $args)));
					}
					else
					{
						self::termPrint(sprintf('#%d %s%s', $i, $file, $line));
					}
				}
			}
		}
		else if (self::$enabled == self::PRODUCTION)
		{
			self::htmlProduction($report);
		}
		else
		{
			// Display debug
			echo self::htmlTemplate(ini_get('error_prepend_string'), $report);

			foreach ($report->errors as $e)
			{
				self::htmlException($e);
			}

			echo '<section><article><h2>Context</h2><table>';

			foreach ($report->context as $name => $value)
			{
				printf('<tr><th>%s</th><td>%s</td></tr>', htmlspecialchars($name), htmlspecialchars($value));
			}

			echo '</table></article></section>';

			echo self::htmlTemplate(ini_get('error_append_string'), $report);
		}

		if (self::$report_auto && self::$report_url)
		{
			self::sendReport($report, self::$report_url);
		}

		if ($exit)
		{
			exit(1);
		}
	}

	/**
	 * Prints a line to STDERR, eventually using a color
	 */
	static public function termPrint($message, $color = null)
	{
		if ($color && self::$term_color)
		{
			$message = chr(27) . $color . $message . chr(27) . "[0m";
		}

		fwrite(STDERR, $message . PHP_EOL);
	}

	/**
	 * Return file location without the document root
	 */
	static protected function getFileLocation($file)
	{
		if (!empty(self::$context['rootDirectory']) && ($pos = strpos($file, self::$context['rootDirectory'])) === 0)
		{
			return '...' . substr($file, strlen(self::$context['rootDirectory']));
		}

		return $file;
	}

	/**
	 * Generates a report from an exception
	 */
	static public function makeReport($e)
	{
		$report = (object) [
			'errors' => [],
		];

		while ($e)
		{
			$class = get_class($e);

			$error = (object) [
				'message'   => $e->getMessage(),
				'errorCode' => $e->getCode(),
				'type'      => in_array($class, ['ErrorException', 'Error']) ? 'PHP error' : $class,
				'backtrace' => [
					(object) [
						'file' => $e->getFile() ? self::getFileLocation($e->getFile()) : null,
						'line' => $e->getLine(),
						'code' => $e->getFile() && $e->getLine() ? self::getSource($e->getFile(), $e->getLine()) : null,
					],
				],
			];

			foreach ($e->getTrace() as $i=>$t)
			{
				// Ignore the error stack from ErrorManager
				if (isset($t['class']) && $t['class'] === __CLASS__ 
					&& ($t['function'] === 'shutdownHandler' || $t['function'] === 'errorHandler'))
				{
					continue;
				}

				$args = [];

				// Display call arguments
				if (!empty($t['args']))
				{
					// Find arguments variables names via reflection
					try {
						if (isset($t['class']))
						{
							$r = new \ReflectionMethod($t['class'], $t['function']);
						}
						else
						{
							$r = new \ReflectionFunction($t['function']);
						}

						$params = $r->getParameters();
					}
					catch (\Exception $e) {
						$params = [];
					}

					foreach ($t['args'] as $name => $value)
					{
						if (array_key_exists($name, $params))
						{
							$name = '$' . $params[$name]->name;
						}

						if (is_string($value))
						{
							$value = self::getFileLocation($value);
						}

						$args[$name] = self::dump($value);

						if (strlen($args[$name]) > 200)
						{
							$args[$name] = substr($args[$name], 0, 199) . '…';
						}
					}
				}

				$trace = (object) [
					// Add class name to function
					'function' => isset($t['class']) ? $t['class'] . $t['type'] . $t['function'] : $t['function'],
				];

				if (isset($t['file']))
				{
					$trace->file = self::getFileLocation($t['file']);
				}

				if (isset($t['line']))
				{
					$trace->line = (int) $t['line'];
				}

				if (count($args))
				{
					$trace->args = $args;
				}

				if (isset($trace->file) && isset($trace->line))
				{
					$trace->code = self::getSource($t['file'], $t['line']);
				}

				$error->backtrace[] = $trace;
			}

			$report->errors[] = $error;
			$e = $e->getPrevious();
		}

		unset($error, $e, $params, $t);

		$report->context = (object) array_merge([
			'id'       => base_convert(substr(sha1(json_encode($report->errors)), 0, 10), 16, 36),
			'date'     => date(DATE_ATOM),
			'os'       => PHP_OS,
			'language' => 'PHP ' . PHP_VERSION,
			'environment' => self::$enabled == self::DEVELOPMENT ? 'development' : 'production:' . self::$enabled,
			'php_sapi' => PHP_SAPI,
		], self::$context);

		if (!empty($_SERVER['HTTP_HOST']) && !empty($_SERVER['REQUEST_URI']))
		{
			$proto = empty($_SERVER['HTTPS']) ? 'http' : 'https';
			$report->context->url = $proto . '://' . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
		}

		return $report;
	}

	/**
	 * Displays an exception as HTML debug page
	 */
	static public function htmlException(\stdClass $e)
	{
		printf('<section><header><h1>%s</h1><h2>%s</h2></header>',
			$e->type, htmlspecialchars($e->message));

		foreach ($e->backtrace as $i=>$t)
		{
			echo '<article>';

			if (isset($t->file) && isset($t->line))
			{
				$dir = dirname($t->file);
				$dir = $dir == '/' ? $dir : $dir . '/';

				printf('<h3>in %s<b>%s</b>:<i>%d</i></h3>', htmlspecialchars($dir), htmlspecialchars(basename($t->file)), $t->line);
			}

			if (isset($t->function))
			{
				printf('<h4>&rarr; %s <i>(%d arg.)</i></h4>', htmlspecialchars($t->function), isset($t->args) ? count($t->args) : 0);

				// Display call arguments
				if (!empty($t->args))
				{
					echo '<table>';

					foreach ($t->args as $name => $value)
					{
						printf('<tr><th>%s</th><td><pre>%s</pre></td>', htmlspecialchars($name), htmlspecialchars($value));
					}

					echo '</table>';
				}
			}

			// Display source code
			if (isset($t->code) && isset($t->line))
			{
				echo self::htmlSource($t->code, $t->line);
			}

			echo '</article>';
		}

		echo '</section>';
	}


	static public function htmlSource(array $source, $line)
	{
		$out = '';

		foreach ($source as $i => $code)
		{
			$html = '<b>' . ($i) . '</b>' . htmlspecialchars($code, ENT_QUOTES);

			if ($i == $line)
			{
				$html = '<u>' . $html . '</u>';
			}

			$out .= $html . PHP_EOL;
		}

		return '<pre><code>' . $out . '</code></pre>';
	}

	/**
	 * Get source code
	 * @param  string $file File location
	 * @param  integer $line Line to highlight
	 * @return array
	 */
	static public function getSource($file, $line)
	{
		$out = [];
		$start = max(0, $line - 5);

		$file = new \SplFileObject($file);
		$file->seek($start);

		for ($i = $start + 1; $i < $start+10; $i++)
		{
			if ($file->eof())
			{
				break;
			}

			$out[$i] = trim($file->current(), "\r\n");
			$file->next();
		}

		unset($file);

		return $out;
	}

	static public function htmlProduction(\stdClass $report)
	{
		if (!headers_sent())
		{
			header('HTTP/1.1 500 Internal Server Error', true, 500);
		}

		echo self::htmlTemplate(self::$production_error_template, $report);
	}

	static public function htmlTemplate($str, \stdClass $report)
	{
		$str = strtr($str, [
			'{$ref}' => $report->context->id,
			'{$report_json}' => htmlspecialchars(json_encode($report), ENT_QUOTES),
			'{$report_url}' => htmlspecialchars(self::$report_url),
		]);

		$str = preg_replace_callback('!<if\((sent|logged|report|email|log)\)>(.*?)</if>!is', function ($match) {
			switch ($match[1]) {
				case 'sent':
				case 'email':
					return self::$email_errors || (self::$report_auto && self::$report_url) ? $match[2] : '';
				case 'logged':
				case 'log':
					return ini_get('error_log') ? $match[2] : '';
				case 'report':
					return (!self::$report_auto && self::$report_url) ? $match[2] : '';
			}
		}, $str);

		return $str;
	}

	/**
	 * Enable error manager
	 * @param  integer $type Type of error management (ErrorManager::PRODUCTION or ErrorManager::DEVELOPMENT)
	 * @return void
	 */
	static public function enable($type = self::DEVELOPMENT)
	{
		if (self::$enabled)
			return true;

		self::$enabled = $type;

		self::$term_color = function_exists('posix_isatty') && @posix_isatty(STDOUT);

		ini_set('display_errors', false);
		ini_set('log_errors', false);
		ini_set('html_errors', false);
		error_reporting($type == self::DEVELOPMENT ? -1 : E_ALL & ~E_DEPRECATED & ~E_STRICT);

		if ($type == self::DEVELOPMENT && PHP_SAPI != 'cli')
		{
			self::setHtmlHeader('<!DOCTYPE html><meta charset="utf-8" /><style type="text/css">
			body { font-family: sans-serif; } * { margin: 0; padding: 0; }
			u, code b, i, h3 { font-style: normal; font-weight: normal; text-decoration: none; }
			#icn { color: #fff; font-size: 2em; float: right; margin: 1em; padding: 1em; background: #900; border-radius: 50%; }
			section header { background: #fdd; padding: 1em; }
			section article { margin: 1em; }
			section article h3, section article h4 { font-size: 1em; font-family: mono; }
			code { border: 1px dotted #ccc; display: block; }
			code b { margin-right: 1em; color: #999; }
			code u { background: #fcc; display: inline-block; width: 100%; }
			table { border-collapse: collapse; margin: 1em; } td, th { border: 1px solid #ccc; padding: .2em .5em; text-align: left; 
			vertical-align: top; }
			</style>
			<pre id="icn"> \__/<br /> (xx)<br />//||\\\\</pre>');
		}

		register_shutdown_function([self::class, 'shutdownHandler']);

		set_exception_handler([__CLASS__, 'exceptionHandler']);

		set_error_handler([__CLASS__, 'errorHandler']);

		if ($type == self::DEVELOPMENT)
		{
			self::startTimer('_global');
		}

		// Assign default context
		static $defaults = [
			'hostname'      => 'SERVER_NAME',
			'userAgent'     => 'HTTP_USER_AGENT',
			'userAddr'      => 'REMOTE_ADDR',
			'remoteAddr'    => 'SERVER_ADDR',
			'rootDirectory' => 'DOCUMENT_ROOT',
		];

		foreach ($defaults as $a => $b)
		{
			if (isset($_SERVER[$b]) && !isset(self::$context[$a]))
			{
				self::$context[$a] = $_SERVER[$b];
			}
		}
	}

	/**
	 * Reset error management to PHP defaults
	 * @return boolean
	 */
	static public function disable()
	{
		self::$enabled = false;

		ini_set('error_prepend_string', null);
		ini_set('error_append_string', null);
		ini_set('log_errors', false);
		ini_set('display_errors', false);
		ini_set('error_reporting', E_ALL & ~E_DEPRECATED & ~E_STRICT);

		restore_error_handler();
		return restore_exception_handler();
	}

	/**
	 * Sets a microsecond timer to track time and memory usage
	 * @param string $name Timer name
	 */
	static public function startTimer($name)
	{
		self::$run_trace[$name] = [microtime(true), memory_get_usage()];
	}

	/**
	 * Stops a timer and return time spent and memory used
	 * @param string $name Timer name
	 */
	static public function stopTimer($name)
	{
		self::$run_trace[$name] = [
			microtime(true) - self::$run_trace[$name][0],
			memory_get_usage() - self::$run_trace[$name][1],
		];
		return self::$run_trace[$name];
	}

	/**
	 * Sets a log file to record errors
	 * @param string $file Error log file
	 */
	static public function setLogFile($file)
	{
		ini_set('log_errors', true);
		return ini_set('error_log', $file);
	}

	/**
	 * Sets an email address that should receive the logs
	 * Set to FALSE to disable email sending (default)
	 * @param string $email Email address
	 */
	static public function setEmail($email)
	{
		self::$email_errors = $email;
	}

	/**
	 * @deprecated
	 */
	static public function setExtraDebugEnv($env)
	{
		self::setContext($env);
	}

	/**
	 * Set the report context
	 * @param mixed $env Variable content, could be application version, or an array of information...
	 */
	static public function setContext(array $context)
	{
		self::$context = $context;
	}

	/**
	 * Enable or disable reporting of errors to a remote URL
	 * @param null|string $url Reporting URL
	 * @param boolean $auto Automatic reporting? If not users will be able to report the error by clicking a button on the error page
	 */
	static public function setRemoteReporting($url, $auto)
	{
		self::$report_url = empty($url) ? null : $url;
		self::$report_auto = (bool) $auto;
	}

	/**
	 * Set the HTML header used by the debug error page
	 * @param string $html HTML header
	 */
	static public function setHtmlHeader($html)
	{
		ini_set('error_prepend_string', $html);
	}

	/**
	 * Set the HTML footer used by the debug error page
	 * @param string $html HTML footer
	 */
	static public function setHtmlFooter($html)
	{
		ini_set('error_append_string', $html);
	}

	/**
	 * Set the content of the HTML template used to display an error in production
	 * {$ref} will be replaced by the error reference if log or email is enabled
	 * <if(email)>...</if> block will be removed if email reporting is disabled
	 * <if(log)>...</if> block will be removed if log reporting is disabled
	 * @param string $html HTML template
	 */
	static public function setProductionErrorTemplate($html)
	{
		self::$production_error_template = $html;
	}

	static public function setCustomExceptionHandler($class, Callable $callback)
	{
		self::$custom_handlers[$class] = $callback;
	}

	/**
	 * Copy of var_dump but returns a string instead of a variable
	 * @param  mixed  $var   variable to dump
	 * @param  integer $level Indentation level (internal use)
	 * @return string
	 */
	static public function dump($var, $level = 0)
	{
		switch (gettype($var))
		{
			case 'boolean':
				return 'bool(' . ($var ? 'true' : 'false') . ')';
			case 'integer':
				return 'int(' . $var . ')';
			case 'double':
				return 'float(' . $var . ')';
			case 'string':
				return 'string(' . strlen($var) . ') "' . $var . '"';
			case 'NULL':
				return 'NULL';
			case 'resource':
				return 'resource(' . (int)$var . ') of type (' . get_resource_type($var) . ')';
			case 'array':
			case 'object':
				if (is_object($var))
				{
					$out = 'object(' . get_class($var) . ') (' . count((array) $var) . ') {' . PHP_EOL;
				}
				else
				{
					$out = 'array(' . count((array) $var) . ') {' . PHP_EOL;
				}

				$level++;

				foreach ($var as $key=>$value)
				{
					$out .= str_repeat(' ', $level * 2);
					$out .= is_string($key) ? '["' . $key . '"]' : '[' . $key . ']';
					$out .= '=> ' . self::dump($value, $level) . PHP_EOL;
				}

				$out .= str_repeat(' ', --$level * 2) . '}';
				return $out;
			default:
				return gettype($var);
		}
	}

	/**
	 * Upload a report to a remote errbit-compatible API
	 * @see https://airbrake.io/docs/api/#create-notice-v3
	 */
	static public function sendReport(\stdClass $report, $url)
	{
		$data = json_encode($report);

		$headers = [
			'Content-Type: application/json',
			'Content-Lenth: ' . strlen($data),
		];

		if (function_exists('curl_init'))
		{
			$ch = curl_init($url);

			curl_setopt_array($ch, [
				CURLOPT_HTTPHEADER     => $headers,
				CURLOPT_RETURNTRANSFER => true,
				CURLOPT_FOLLOWLOCATION => false,
				CURLOPT_MAXREDIRS      => 3,
				CURLOPT_CUSTOMREQUEST  => 'POST',
				CURLOPT_TIMEOUT        => 10,
				CURLOPT_POSTFIELDS     => $data,
			]);

			$body = curl_exec($ch);
			$code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		}
		else
		{
			$opts = ['http' => [
				'method'        => 'POST',
				'header'        => $headers,
				'content'       => $data,
				'max_redirects' => 3,
				'timeout'       => 10,
				'ignore_errors' => true,
			]];

			$body = file_get_contents($url, false, stream_context_create($opts));
			$code = null;

			foreach ($http_response_header as $header)
			{
				$a = substr($header, 0, 7);

				if ($a == 'HTTP/1.')
				{
					$code = substr($header, 11, 3);
				}
			}

			unset($http_response_header);
		}

		return [
			'code' => (int) $code,
			'body' => $body,
			'data' => json_decode($body),
		];
	}

	/**
	 * Returns list of reports from error log
	 *
	 * @param string|null $log_file Log file to use, if NULL then the log file set in error_log will be used
	 * @param string|null $filter_id Only return errors matching with this ID
	 */
	static public function getReportsFromLog($log_file = null, $filter_id = null)
	{
		if (!$log_file)
		{
			$log_file = ini_get('error_log');
		}

		if (!file_exists($log_file))
		{
			return $log_file;
		}

		$reports = [];
		$report = null;

		foreach (file($log_file) as $line)
		{
			$line = trim($line);

			if ($line == '<errorReport>')
			{
				$report = '';
			}
			elseif ($line == '</errorReport>')
			{
				$report = json_decode($report);

				if (!is_null($report) && isset($report->context->id) && (!$filter_id || $filter_id == $report->context->id))
				{
					$reports[] = $report;
				}

				$report = null;
			}
			elseif ($report !== null)
			{
				$report .= $line;
			}
		}

		unset($line, $report, $log_file);

		return $reports;
	}
}
